package middleware

import (
	"context"

	pkgId "git.perx.ru/perxis/perxis-go/id"
	"git.perx.ru/perxis/perxis-go/pkg/locales"
	logzap "git.perx.ru/perxis/perxis-go/zap"
	"go.uber.org/zap"
)

type loggingMiddleware struct {
	logger *zap.Logger
	next   locales.Locales
}

func LoggingMiddleware(logger *zap.Logger) Middleware {
	return func(next locales.Locales) locales.Locales {
		return &loggingMiddleware{
			next:   next,
			logger: logger.With(logzap.Component("Locales")),
		}
	}
}

func (m *loggingMiddleware) Create(ctx context.Context, locale *locales.Locale) (created *locales.Locale, err error) {
	logger := m.logger.With(
		logzap.Caller(ctx),
		logzap.Event(locales.EventCreate),
	)

	created, err = m.next.Create(ctx, locale)
	if err != nil {
		logger.Error("Failed to create", zap.Error(err), logzap.Channels(logzap.Userlog, logzap.Syslog), logzap.Object(locale))
		return
	}

	logger.Info("Locale created", logzap.Channels(logzap.Userlog), logzap.Object(created))
	return created, err
}

func (m *loggingMiddleware) Update(ctx context.Context, locale *locales.Locale) (err error) {
	logger := m.logger.With(
		logzap.Caller(ctx),
		logzap.Event(locales.EventUpdate),
		logzap.Object(locale),
	)

	err = m.next.Update(ctx, locale)
	if err != nil {
		logger.Error("Failed to update", zap.Error(err), logzap.Channels(logzap.Userlog, logzap.Syslog))
		return
	}

	logger.Info("Locale updated", logzap.Channels(logzap.Userlog))
	return err
}

func (m *loggingMiddleware) List(ctx context.Context, spaceId string) (locales []*locales.Locale, err error) {
	logger := m.logger.With(
		logzap.Caller(ctx),
	)

	locales, err = m.next.List(ctx, spaceId)
	if err != nil {
		logger.Error("Failed to list", zap.Error(err))
		return
	}

	return locales, err
}

func (m *loggingMiddleware) Delete(ctx context.Context, spaceId, id string) (err error) {
	logger := m.logger.With(
		logzap.Caller(ctx),
		logzap.Event(locales.EventDelete),
		logzap.Object(pkgId.NewLocaleId(spaceId, id)),
	)

	err = m.next.Delete(ctx, spaceId, id)
	if err != nil {
		logger.Error("Failed to delete", zap.Error(err), logzap.Channels(logzap.Userlog, logzap.Syslog))
		return
	}

	logger.Info("Locale deleted", logzap.Channels(logzap.Userlog))
	return err
}
